# Name: xdp/interpolate/bezier.py
# Purpose: Provides bezier curve generation and interpolation
# Author: Ken McIvor <mcivken@iit.edu>
#
# Copyright 2003-2004 Illinois Institute of Technology
#
# See the file "LICENSE" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

"""
Sliding-window interpolation using bezier curves.
"""


import sys
import bisect
import Numeric

import _bezier

__all__ = ['bint']


def bint(xVec, yVec, xRange, order=3):
    """
    Perform a sliding window interpolation using bezier curves.

    @param xVec: independent values
    @type  xVec: C{Numeric} vector
    @param yVec: dependent values
    @type  yVec: C{Numeric} vector
    @param xRange: abscissa
    @type  xRange: C{Numeric} vector
    @param order: order of bezier curves to use (M{order+1} points per curve)
    @type  order: C{int}
    @return: indepolated dependent values
    @rtype:  C{Numeric} vector
    """

    pointcount = order + 1
    coeff = Numeric.zeros( (pointcount,), 'd')
    _bezier.coeff(coeff)
    out = Numeric.zeros( (2,), 'd')

    newY = Numeric.zeros( (len(xRange),), 'd')

    slideStart = pointcount / 2 + pointcount % 2
    slideEnd = xVec.shape[0] - pointcount
    windowEnd = pointcount/2 + pointcount % 2

    for i in range(0, xRange.shape[0]):
        start = bisect.bisect(xVec, xRange[i]) - 1

        if xRange[i] < xVec[start] and (start > slideStart or start < slideEnd):
            while xRange[i] >= xVec[start]:
                start = start + 1

        if start < slideStart:
            start = 0
        elif start > slideEnd:
            start = slideEnd
        else:
            start = start - windowEnd

        xWindow = xVec[start:start+pointcount]
        yWindow = yVec[start:start+pointcount]
        u = (xRange[i] - xWindow[0]) / (xWindow[-1] - xWindow[0])
        _bezier.bint(u, xWindow, yWindow, coeff, out)
        newY[i] = out[1]
    return newY
