# Name: xdp/io/__init__.py
# Purpose: XDP I/O package
# Author: Ken McIvor <mcivor@iit.edu>
#
# Copyright 2003-2005 Illinois Institute of Technology
#
# See the file "LICENSE" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

"""Read and write experiment data and metadata.

The file formats supported by C{xfp}:
 - MX SFF S{->} L{Sff}, L{readSffFile}
 - MR-CAT XAFS 0.3 S{->} L{Xafs}, L{readXafsFile}
 - Generic Text S{->} L{Text}, L{readTextFile}
"""


import os
import sre

from sff import SffFieldError, Sff, readSffFile
from xafs import Xafs, readXafsFile
from text import Text, readTextFile


__all__ = ['SffFieldError', 'Sff', 'readSffFile', 'Xafs', 'readXafsFile',
    'Text', 'readTextFile', 'readFile', 'writeFile']


def readFile(fileName):
    """Read experiment data and metadata from a file.

    The format of the file is guessed from the file's extension and
    the contents.

    @param fileName: file to open
    @type  fileName: C{str}ing

    @return: header and data
    @rtype:  C{2-tuple} of the form C{(header, data)}

    @raise IOError: could not guess the file format, or open, read, or parse
                    the file

    @todo: verify that the file access mode does not need to be binary
    """

    opener = _guessFormat(fileName)
    if not opener:
        raise IOError('could not guess format of %s' % fileName)

    input = open(fileName, 'r')
    return opener(input)


def writeFile(fileName, header, data):
    """
    Write experiment data and metadata to a file.

    The format of the file is determined by the class of the C{header} argument.

    @param fileName: file to write
    @type  fileName: C{str}ing

    @param header: metadata
    @type  header: C{SFF} or C{Xafs} or C{Text}

    @param data: dataset
    @type  data: C{Data}

    @raise IOError: either C{header} or C{data} contain invalid information, or
                    the file could not be opened for writing

    @todo: verify that the file access mode does not need to be binary
    """

    output = open(fileName, 'w')
    header.write(data, output)


#
# Implementation Functions
#

FMT_SFF  = (sre.compile(r'^(#\s*)?SFF'),         readSffFile)
FMT_TEXT = (sre.compile(r'^\s*(#|!|;|-|\+|\.|[0-9])'), readTextFile)
FMT_XAFS = (sre.compile(r'^MRCAT_XAFS'),         readXafsFile)


_FORMATS = {
    'dat':  FMT_TEXT,
    'sff':  FMT_SFF,
    'txt':  FMT_TEXT,
    'text': FMT_TEXT,
    'xafs': FMT_XAFS,
}

_FORMAT_LIST = 'sff', 'xafs', 'text'


def _guessFormat(path):
    """
    Attempt to determine the file format.  C{None} is returned if the format is
    unknown.

    @param path: file
    @type  path: C{str}ing
    @return: input function
    @rtype:  C{readXXXFile} fuction or C{None}
    """

    filename, format = os.path.splitext(os.path.basename(path))
    format = format[1:].lower()

    if format and _FORMATS.has_key(format):
        return _FORMATS[format][1]

    input = open(path, 'r')
    line = input.read(20)
    input.close()

    for regexp, func in [_FORMATS[x] for x in _FORMAT_LIST]:
        if regexp.match(line):
            return func

    return None
