# Name: xdp/io/sff.py
# Purpose: I/O support for Simple File Format (SFF) data files
# Author: Ken McIvor <mcivor@iit.edu>
#
# Copyright 2003-2005 Illinois Institute of Technology
#
# See the file "LICENSE" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

"""
Read and write files using Mx's Simple File Format.
"""


import types
import string
import Numeric
import operator

import xdp
import _matrixio


__all__ = ['SffFieldError', 'Sff', 'readSffFile']


def SffFieldError(Exception):
    """
    Exception raised to indicate that an invalid field name was specified.
    """
    pass


class Sff:
    """
    The contents of the header of an SFF file.
    """

    def __init__(self, **kwds):
        """Create an empty SFF header.

        If keyword arguments are provided, then are interpreted as field
        definitions and are used to populate the header.

        @param kwds: field definitions
        @type  kwds: C{dict}
        """
        self.__fields = kwds

    def fields(self):
        """
        Retrive a list of the field names.

        @return: names of this header's fields
        @rtype:  C{list} of C{str}ings

        @depreciated: use C{getFieldNames} instead
        """

        return self.__fields.keys()

    def getFieldNames(self):
        """
        Retrive a list of the field names.

        @return: names of this header's fields
        @rtype:  C{list} of C{str}ings
        """

        return self.__fields.keys()

    def has(self, name):
        """
        Check for the presence of a field.

        @param name: field to check for
        @type  name: C{str}ing
        @return: whether or not this dataset has a field C{name}
        @rtype:  C{bool}

        @depreciated: use C{hasField} instead
        """

        return self.__fields.has_key(name)

    def hasField(self, name):
        """
        Check for the presence of a field.

        @param name: field to check for
        @type  name: C{str}ing
        @return: whether or not this dataset has a field C{name}
        @rtype:  C{bool}

        @depreciated: use C{hasField} instead
        """

        return self.__fields.has_key(name)

    def get(self, name):
        """
        Retrieve a field's value.

        @param name: field name
        @type  name: C{str}ing
        @return: field value
        @rtype:  I{any}

        @raise SffFieldError: this header has no field C{name}

        @depreciated: use C{getField} instead
        """

        if self.__fields.has_key(name):
            return self.__fields[name]
        else:
            raise SffFieldError(name)

    def getField(self, name):
        """
        Retrieve a field's value.

        @param name: field name
        @type  name: C{str}ing
        @return: field value
        @rtype:  I{any}

        @raise SffFieldError: this header has no field C{name}
        """

        if self.__fields.has_key(name):
            return self.__fields[name]
        else:
            raise SffFieldError(name)

    def add(self, name, value):
        """
        Create a new field.

        @param name: new field's name
        @type  name: C{str}ing
        @param value: new field's value
        @type  value: I{any}

        @raise SffFieldError: the field C{name} already exists

        @depreciated: use C{addField} instead
        """

        if not self.__fields.has_key(name):
            self.__fields[name] = value
        else:
            raise SffFieldError(name)

    def addField(self, name, value):
        """
        Create a new field.

        @param name: new field's name
        @type  name: C{str}ing
        @param value: new field's value
        @type  value: I{any}

        @raise SffFieldError: the field C{name} already exists
        """

        if not self.__fields.has_key(name):
            self.__fields[name] = value
        else:
            raise SffFieldError(name)

    def set(self, name, value):
        """
        Assign a value to a field.  If the field does not exist, it will be
        created.

        @param name: field name
        @type  name: C{str}ing
        @param value: new field data
        @type  value: I{any}

        @depreciated: use C{setField} instead
        """

        self.__fields[name] = value

    def setField(self, name, value):
        """
        Assign a value to a field.  If the field does not exist, it will be
        created.

        @param name: field name
        @type  name: C{str}ing
        @param value: new field data
        @type  value: I{any}
        """

        self.__fields[name] = value

    def remove(self, name):
        """
        Remove a field from the header.

        @param name: field to remove
        @type  name: C{str}ing

        @raise SffFieldError: the field C{name} does not exist

        @depreciated: use C{removeField} instead
        """

        if self.__fields.has_key(name):
            del self.__fields[name]
        else:
            raise SffFieldError(name)

    def removeField(self, name):
        """
        Remove a field from the header.

        @param name: field to remove
        @type  name: C{str}ing

        @raise SffFieldError: the field C{name} does not exist
        """

        if self.__fields.has_key(name):
            del self.__fields[name]
        else:
            raise SffFieldError(name)

    def write(self, data, stream):
        """
        Write a dataset to an output stream in SFF, using this header's
        fields.

        The output is written in the SFF format, and a I{%devices} field is
        created to hold the column names from C{data}.

        @param data: dataset
        @type  data: C{Data}
        @param stream: output stream
        @type  stream: C{file}
        """

        header = self.__fields.copy()
        header['%devices'] = string.join(data.columns(), ' ')

        fields = header.keys()
        fields.sort()

        stream.write('# SFF main_header %d\n' % len(fields))

        for fieldName in fields:
            value = header[fieldName]

            if isinstance(value, types.StringType):
                stream.write('# %s = "%s";\n' % (fieldName, repr(value)[1:-1]))
            elif isinstance(value, types.UnicodeType):
                value = value.encode('ASCII')
                stream.write('# %s = "%s";\n' % (fieldName, repr(value)[1:-1]))
            elif operator.isSequenceType(value):
                value = [ str(x) for x in value ]
                value = string.join(value, ' ')
                stream.write('# %s = %s;\n' % (fieldName, value))
            else:
                stream.write('# %s = %s;\n' % (fieldName, value))

        _matrixio.write(stream, data.getMatrix())


def readSffFile(stream):
    """
    Read an SFF header and data from an input stream.

    @param stream: input stream
    @type  stream: C{file}
    @return: header and data
    @rtype:  C{2-tuple} of the form C{(header, data)}

    @raise IOError: the SFF header or data was malformed

    @note: Only SFF files that have a single header are supported.
    """

    type, headerDict = _sff_readHeader(stream)

    if type != 'main_header':
        raise IOError('main header has incorrect type')

    try:
        columnNames = headerDict['%devices']
    except KeyError:
        columnNames = []
    else:
        del headerDict['%devices']

    header = Sff(**headerDict)
    data = xdp.Data()

    try:
        dataMatrix = _matrixio.read(stream)
    except _matrixio.error, e:
        raise IOError(str(e))

    if not columnNames:
        columnNames = ['column%s' % x for x in range(1, dataMatrix.shape[1]+1)]

    nCols = dataMatrix.shape[1]
    nNames = len(columnNames)

    if nNames < nCols:
        N = nNames
        extraCols = nCols - nNames
    else:
        N = nCols
        extraCols = 0

    for i in range(0, N):
        data.add(columnNames[i], dataMatrix[:, i])

    for i in range(N, extraCols + N):
        data.add('column%d' % i, dataMatrix[:, i])

    return header, data


#
# Private implementation functions
#


def _sff_read_header_line(stream):
    """Read single line of an SFF header from an input stream.

    @param stream: input stream
    @type  stream: C{file}
    @return: the contents of the next header line of the stream
    @rtype:  C{string}

    @raise IOError: end-of-file was encountered
    """

    line = stream.readline()
    if line.startswith('#'):
        line = line[1:].lstrip()

    if not line:
        raise IOError('unexpected end of file')

    line = line.strip()
    if not line:
        raise IOError('empty line in SFF header')

    return line


def _sff_readHeader(stream):
    """Read an SFF header from an input stream.

    @param stream: input stream
    @type  stream: C{file}
    @return: header type and fields
    @rtype:  C{2-tuple} of the form C{(header_type, fields_dict)}

    @raise IOError: the SFF header was missing or malformed
    """

    line = _sff_read_header_line(stream)

    if line[0:3] != 'SFF':
        raise IOError('sff header line missing')

    pieces = string.split(line)
    if len(pieces) != 3:
        raise IOError('malformed header')
    else:
        headerType, fieldCount = pieces[1:3]
        try:
            fieldCount = int(fieldCount)
        except ValueError:
            raise IOError('malformed header line count')

    fields = {}
    for i in range(0, fieldCount):
        line = _sff_read_header_line(stream)
        _sff_parseHeaderField(line, fields)
    return headerType, fields


def _sff_parseHeaderField(line, fields):
    """Parse one SFF header field from a line.

    The I{%devices} field is split by whitespace into a list of the column
    names.  Other fields that being with a percent sign are not modified.

    The field name and value are stored in C{fields}.

    @param line: one line from an SFF header
    @type  line: C{str}ing
    @param fields: field names and values
    @type  fields: C{dict}

    @raise IOError: the SFF header is missing or malformed

    @todo: refactor this function
    """

    pieces = string.split(line, '=', 1)
    if len(pieces) != 2:
        raise IOError('malformed field')
    else:
        name = string.strip(pieces[0])
        value = string.strip(pieces[1])
        if value[-1] != ';':
            raise IOError('field "%s" is unterminated' % name)
        else:
            value = value[0:-1]

    if name == '%devices':
        if value[0] == '"':
            if value[-1] != '"':
                raise IOError('unterminated string value in field "%s"' % name)
            else:
                value = value[1:-1]
        value = string.split(value)
    elif name[0] == '%':
        pass
    elif value[0] == '"':
        if value[-1] != '"':
            raise IOError('unterminated string value in field "%s"' % name)
        else:
            value = value[1:-1]
    else:
        try:
            value = int(value)
        except ValueError:
            try:
                value = long(value)
            except ValueError:
                try:
                    value = float(value)
                except ValueError:
                    pass

    if fields.has_key(name):
        raise IOError('field name "%s" is a duplicate' % name)
    else:
        fields[name] = value
