/* Name: _matrixio.c
 * Purpose: fast input/output for two dimensional Numerical Python arrays
 * Author: Ken McIvor <mcivken@iit.edu>
 *
 * Copyright 2003-2004 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <Python.h>
#include <Numeric/arrayobject.h>
#define MATRIXIO_MODULE
#include "_matrixio.h"


static char *__version__ = "1.1";
static char *__rcsid__ = "$Id$";
static char *__source__ = "$Source$";


static PyObject *mioError;
#define pyRaise(type, mesg) { PyErr_SetString(type, mesg); return NULL; }


static PyObject *matrixio_pyRead(PyObject *self, PyObject *args)
{
    PyObject *fObj;
    double *data;
    FILE *iFile;
    int dims[2];


    if (!PyArg_ParseTuple(args, "O!", &PyFile_Type, &fObj))
        return NULL;

    Py_INCREF(fObj);

    iFile = PyFile_AsFile(fObj);
    if (iFile == NULL) {
        Py_DECREF(fObj);
        pyRaise(PyExc_IOError, \
            "failed to obtain FILE pointer via PyFile_AsFile()");
    }

    switch (matrixio_read(iFile, &data, &dims[0], &dims[1])) {
    case _MERR_OKAY:
        break;

    case _MERR_IBUFFMEM:
        pyRaise(PyExc_MemoryError, "could not allocate an input buffer");

    case _MERR_BUFFMEM:
        pyRaise(PyExc_MemoryError, "could not resize the input buffer");

    case _MERR_IDATAMEM:
        pyRaise(PyExc_MemoryError, "could not allocate a data array");

    case _MERR_DATAMEM:
        pyRaise(PyExc_MemoryError, "could not resize the data array");

    case _MERR_CONVERT:
        pyRaise(mioError, "row contains an invalid element");

    case _MERR_FEWCOL:
        pyRaise(mioError, "row contains too few columns");

    case _MERR_MANYCOL:
        pyRaise(mioError, "row contains too many columns");

    default:
        pyRaise(mioError, "unknown return value for matrixio_read()");
    }

    Py_DECREF(fObj);
    return PyArray_FromDimsAndData(2, dims, PyArray_DOUBLE, (char *) data);
}


static PyObject *matrixio_pyWrite(PyObject *self, PyObject *args)
{
    PyArrayObject *array;
    PyObject *fObj;
    FILE *oFile;

    if (!PyArg_ParseTuple(args, "O!O!", &PyFile_Type, &fObj, &PyArray_Type, \
       &array))
        return NULL;

    if (array->nd != 2)
        pyRaise(PyExc_ValueError, "array must be two dimensional");
    
    if (array->descr->type_num != PyArray_DOUBLE)
        pyRaise(PyExc_ValueError, "array must be of type Numeric.Float64");

    oFile = PyFile_AsFile(fObj);
    if (oFile == NULL) {
        Py_DECREF(fObj);
        pyRaise(PyExc_IOError, \
            "failed to obtain FILE pointer via PyFile_AsFile()");
    }
 
    /* FIXME: should check for errors and raise an IOError */
    matrixio_write(oFile, (PyArrayObject *) array);

    Py_INCREF(Py_None);
    return Py_None;
}


/*
 * TODO: track line number to provide better diagnostic information
 */
static int matrixio_read(FILE *iFile, double **data, int *rows, int *columns)
{
    char *c, *next, *endptr,
        *buff = NULL;
    double elem;
    int len, nc,
        elemcount = 0,
        countedColumns = 0,
        buffsize = _MIO_BUFFSTART,
        datasize = _MIO_DATASTART;

    *rows = *columns = 0;

    buff = (char *) malloc(_MIO_BUFFSTART);
    if (buff == NULL)
        return _MERR_IBUFFMEM;

    *data = (double *) malloc(_MIO_DATASTART * sizeof(double));
    if (data == NULL)
        return _MERR_IDATAMEM;

    while (1) {
        if (fgets(buff, buffsize, iFile) == NULL)
            break;

        while (1) {
            len = strlen(buff);
            if (buffsize == len+1 && buff[len-1] != '\n') {
                buffsize += _MIO_BUFFINCR;
                buff = realloc(buff, buffsize);
                if (buff == NULL) {
                    free(*data);
                    return _MERR_BUFFMEM;
                }
                fgets(buff + len, buffsize - len, iFile);
            } else {
                break;
            }
        }

        nc = 0;
        next = buff;
        while (next != NULL) {
            /* FIXME: address portability issues (i.e. reading a CRLF file
             * in linux, reading an LF file in windows, etc) */
            c = strsep(&next, " \t");
            while (next && *next && (*next == ' ' || *next == '\t'))
                ++next;

            if (*c == '\0' || *c == ' ' || *c == '\t' || *c == '\n' || *c =='\r') {
                continue;
            }

            elem = strtod(c, &endptr);
            if (endptr == c) {
                fprintf(stderr, "Could not convert \"%s\"\n", c);
                free(buff);
                free(*data);
                return _MERR_CONVERT;
            }

            (*data)[elemcount++] = elem;

            if (elemcount == datasize) {
                datasize += _MIO_DATAINCR;
                *data = realloc(*data, datasize * sizeof(double));
                if (*data == NULL) {
                    free(buff);
                    return _MERR_DATAMEM;
                }
            }

            ++nc;
            if (countedColumns) {
                if (nc > *columns) {
                    errno = EBADF;
                    free(buff);
                    free(*data);
                    return _MERR_MANYCOL;
                }
            } else {
                *columns = nc;
            }
        }

        if (nc < *columns) {
            errno = EBADF;
            free(buff);
            free(*data);
            return _MERR_FEWCOL;
        }

        if (countedColumns == 0)
            countedColumns = 1;
        ++(*rows);
    }

    free(buff);
    return 0;
}


static int matrixio_write(FILE *oFile, PyArrayObject *array)
{
    int r, c;

    if (array->nd != 2 || array->descr->type_num != PyArray_DOUBLE)
        return -1;

    for (r = 0; r < array->dimensions[0]; r++) {
        for (c = 0; c < array->dimensions[1]; c++) {
            /* FIXME: should check for errors */
            fprintf(oFile, "%f\t", *(double *)(array->data + \
                r*array->strides[0] + c*array->strides[1]));
        }
        /* FIXME: should check for errors */
        fputs("\n", oFile);
    }

    return 0;
}


static PyMethodDef matrixioMethods[] = {
    {"read", matrixio_pyRead, METH_VARARGS,
        "read(fileObj) -> Numerical array -- reads a matrix from fileObj"},
    {"write", matrixio_pyWrite, METH_VARARGS,
        "write(fileObj, matrix) -- writes matrix to fileObj"},
    {NULL, NULL, 0, NULL}
};


void init_matrixio(void)
{
    PyObject *m, *s, *dict;
    static void *matrixio_API[_matrixio_API_pointers];
    PyObject *c_api_object;

    m = Py_InitModule("_matrixio", matrixioMethods);

    matrixio_API[_matrixio_pyRead_NUM] = (void *) matrixio_pyRead;
    matrixio_API[_matrixio_pyWrite_NUM] = (void *) matrixio_pyWrite;
    matrixio_API[_matrixio_read_NUM] = (void *) matrixio_read;
    matrixio_API[_matrixio_write_NUM] = (void *) matrixio_write;
    c_api_object = PyCObject_FromVoidPtr((void *) matrixio_API, NULL);

    if (c_api_object != NULL) {
        PyObject *d = PyModule_GetDict(m);

        PyDict_SetItemString(d, "_C_API", c_api_object);
        Py_DECREF(c_api_object);
    }

    dict = PyModule_GetDict(m);
    mioError = PyErr_NewException("xdp.io._matrixio.error", NULL, NULL);
    if (mioError != NULL)
        PyDict_SetItemString(dict, "error", mioError);

    s = Py_BuildValue("s", __version__);
    if (s != NULL)
        PyDict_SetItemString(dict, "__version__", s);

    s = Py_BuildValue("s", __rcsid__);
    if (s != NULL)
        PyDict_SetItemString(dict, "__rcsid__", s);

    s = Py_BuildValue("s", __source__);
    if (s != NULL)
        PyDict_SetItemString(dict, "__source__", s);

    import_array();

    if (PyErr_Occurred())
        Py_FatalError("can't initialize module xdp.io._matrixio");
}
